/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#ifndef __CLASS_H_
#define __CLASS_H_

#include "common.h"

#define CLASS_HEAD                              \
    const char* name;                           \
    unsigned int size;                          \
    void (*init) (void* self);                  \
    void (*fini) (void* self)

#define OBJECT_HEAD(classtype)                  \
    classtype* class;                           \
    int refcnt                                  \

#define CALLMETH(obj, name, ...) (obj)->class->name((void*)(obj), ## __VA_ARGS__)

#define METHOD(rt, class, name, ...)                            \
rt class##_##name (class* self, ## __VA_ARGS__);                \
static void CTOR _init_method_ ## class ## _ ## name (void) {   \
    class##_class.name =                                        \
        (typeof(class##_class.name))class##_##name;             \
}                                                               \
rt class##_##name (class* self, ## __VA_ARGS__)

#define _CLASSDEF(classtype, nm, ...)           \
typedef struct nm {                             \
    OBJECT_HEAD(classtype);                     \
    __VA_ARGS__;                                \
} nm;                                           \
                                                \
extern classtype nm ## _class

#define _INITCLASS(classtype, nm, init, fini, ...)      \
classtype nm ## _class = {                              \
    #nm, sizeof(nm),                                    \
    (void (*)(void*))init,                              \
    (void (*)(void*))fini,                              \
    ## __VA_ARGS__                                      \
};

#define object_new(class) ((class*)_object_new(&class##_class))

typedef struct {
    CLASS_HEAD;
} baseclass;

typedef struct {
    OBJECT_HEAD(baseclass);
} object;


void* _object_new(void* class);
int _default_method(void);
void object_ref(void* obj);
void object_deref(void* obj);


#endif
