/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * miscactors.c: Explosions, doors, keys, etc.
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "miscactors.h"
#include "graphics.h"
#include "board.h"
#include "soundutil.h"
#include "player.h"


TEXGROUP(explosion);
ACTOR_ICLASSDEF(explosion, int time);

METHOD(void, explosion, move) {
    self->time++;
    if (self->time > 8)
	actor_remove((actor *) self);
}

METHOD(void, explosion, draw)  {
    addsprite(self->x, self->y, 0, explosion, self->time, DIR_FACE, 0);
}

SERIALIZER(explosion) {
    S_INT(time);
}

void addexplosion(int x, int y) {
    explosion *e = object_new(explosion);
    e->x = x;
    e->y = y;
    e->ang = 0;
    e->time = 0;
    actor_add(e);
}


struct door;

TEXGROUP(door);
ACTOR_ICLASSDEF(doorsensor, struct door* d);
ACTOR_ICLASSDEF(door, 
                int dir, base;
                int opendist, opentime;
                int canclose, needskey;
                doorsensor *sensor);


SERIALIZER(door) {
    S_INT(dir);
    S_INT(base);
    S_INT(opendist);
    S_INT(opentime);
    S_INT(needskey);
    S_ACTOR(sensor);

    if (phase == PHASE_LINK) {
        self->sensor->d = self;
        self->canclose = 0;
    }
}

METHOD(void, doorsensor, fini) {
    if (self->d) {
	self->d->sensor = NULL;
        self->d = NULL;
        object_deref(self);
    }

    actor_fini(self);
}

METHOD(void, door, fini) {
    if (self->sensor) {
	self->sensor->d = NULL;
	actor_remove(self->sensor);
    }
    object_deref(self->sensor);
    self->sensor = NULL;

    actor_fini(self);
}


METHOD(void, doorsensor, touch, actor* o) {
    door *d = self->d;
    if (!d)
	return;
    if (HASTAG(o, is_player) && (!d->needskey || HASTAG(o, has_key))) {
	if (!d->opentime && !d->opendist) {
	    playsndxy("door", self->x, self->y, 64);
	    d->opentime = 72;
	} else {
	    d->canclose = 0;
	}

    }
}

METHOD(void, door, touch, actor* o) {
    if (o != (actor*)self->sensor) {
	if (self->opendist && !self->opentime) {
	    self->opentime = 24;
	}
	self->canclose = 0;
    }
}


#define DOORSPEED 3500

METHOD(void, door, move) {
    int bx, by;
    int dd;
    int sx, sy;

    if (self->sensor && !self->sensor->d) {
        self->sensor->d = self;
    }

    sx = self->x;
    sy = self->y;
    if (self->dir & 1) {
	sy = self->base;
    } else {
	sx = self->base;
    }
    bx = sx >> 16;
    by = sy >> 16;
    dd = 0;
    if (self->opentime) {
	dd = 65536;
	if (!--self->opentime) {
	    if (self->canclose) {
		BCELL(bx, by) |= 16384;
		playsndxy("door", self->x, self->y, 64);
	    } else {
		self->opentime++;
	    }
	}
    }
    if (self->opendist < dd) {
	self->opendist += DOORSPEED;
	if (self->opendist >= dd) {
	    BCELL(bx, by) &= ~16384;
	    self->opendist = dd;
	}
    }
    if (self->opendist > dd) {
	self->opendist -= DOORSPEED;
	if (self->opendist <= dd) {
	    self->opendist = dd;
	    playsndxy("doorclose", self->x, self->y, 64);
	}

    }
    switch (self->dir & 7) {
    case 0:
	self->w = (65536 - self->opendist) >> 1;
	self->x = self->base + 32768 - self->w;
	break;
    case 2:
	self->w = (65536 - self->opendist) >> 1;
	self->x = self->base - 32768 + self->w;
	break;
    case 1:
	self->h = (65536 - self->opendist) >> 1;
	self->y = self->base + 32768 - self->h;
	break;
    case 3:
	self->h = (65536 - self->opendist) >> 1;
	self->y = self->base - 32768 + self->h;
	break;
    }

    self->canclose = 1;
}

METHOD(void, door, draw) {
    int sx, sy;
    if (self->opendist == 65536)
	return;

    sx = self->x;
    sy = self->y;
    if (self->dir & 1) {
	sy = self->base;
    } else {
	sx = self->base;
    }

    switch (self->dir&3) {
    case 0:
	sx += self->opendist;
	break;
    case 1:
	sy += self->opendist;
	break;
    case 2:
	sx -= self->opendist;
	break;
    case 3:
	sy -= self->opendist;
	break;
    }
    
    addsprite(sx, sy, 0, door, self->needskey, self->dir & 1 ? DIR_WEST : DIR_NORTH, 0);
    addsprite(sx, sy, 0, door, self->needskey, self->dir & 1 ? DIR_EAST : DIR_SOUTH, 0);
}

METHOD(int, door, clips) {
    return self->opendist < 65536;
}

METHOD(void, door, init) {
    actor_init(self);

    self->w = self->h = 32768;
}

METHOD(void, doorsensor, init) {
    actor_init(self);

    self->w = self->h = 32768;
}

METHOD(void, door, load, buffer* b) {
    actor_load(self, b);
    self->opentime = reads16(b);
    self->opendist = readu32(b);
    self->dir = readu8(b);
    self->needskey = readu8(b);
    if (self->dir & 1) {
	self->w = 2;
	self->base = self->y;
    } else {
	self->h = 2;
	self->base = self->x;
    }
    self->sensor = object_new(doorsensor);
    self->sensor->x = self->x;
    self->sensor->y = self->y;
    self->sensor->ang = 0;

    object_ref(self->sensor);
    if (self->sensor) {
	if (self->dir & 1) {
	    self->sensor->w = 24576;
	} else {
	    self->sensor->h = 24576;
	}
	self->sensor->d = self;
	actor_add(self->sensor);
    }

    BCELL(self->x >> 16, self->y >> 16) |= 16384;
}


/*
 *  GENERATOR
 */
TEXGROUP(generator);
ACTOR_ICLASSDEF(generator,
	  buffer dat;
	  
	  actor* creating;
	  
	  int create_timer;
	  
	  int timer;
	  int numactors;
	  int min_interval;
	  int max_interval);


SERIALIZER(generator) {
    S_BUFFER(dat);
    S_ACTOR(creating);
    S_INT(create_timer);
    S_INT(timer);
    S_INT(numactors);
    S_INT(min_interval);
    S_INT(max_interval);
}

METHOD(void, generator, load, buffer* b) {
    actor_load(self, b);
    self->min_interval = readu16(b);
    self->max_interval = readu16(b);
    self->timer = randrange(self->min_interval, self->max_interval);

    self->numactors = readu16(b);
    self->dat.len = readu16(b);
    self->dat.data = malloc(self->dat.len);
    RB memcpy(self->dat.data, &b->data[b->pos], self->dat.len);
    b->pos += self->dat.len;
}

METHOD(void, generator, init) {
    actor_init(self);

    self->dat.data = NULL;
    self->creating = NULL;
}
METHOD(void, generator, fini) {
    object_deref(self->creating);
    self->creating = NULL;
    if (self->dat.data)
        free(self->dat.data);
    self->dat.data = NULL;

    actor_fini(self);
}

HOOK(generator_destroy) {
    generator* self = (generator*)h->params[0].act;
    self->numactors++;
}

#define GENERATOR_TIME 12

METHOD(void, generator, move) {
    int bx, by, i;
    if (self->creating) {
	self->create_timer--;
	if (self->create_timer <= 0) {
	    if (!actor_add(self->creating)) {
                object_deref(self->creating);
		self->numactors++;
	    }
	    self->creating = NULL;
	}
    }
    if (!self->numactors) return;

    findgoal(cplayer->x >> 16, cplayer->y >> 16);

    bx = self->x >> 16;
    by = self->y >> 16;

    i = FINDBUF(bx, by);

    if (i < 24) {
	if (self->timer) {
	    self->timer--;
	    if (!self->timer) {
		actor* o;
		self->timer = randrange(self->min_interval, self->max_interval);
		self->numactors--;
		self->dat.pos = 0;
		o = actor_create(&self->dat);
		if (o) {
		    self->creating = o;
		    self->create_timer = GENERATOR_TIME;
		    playsndxy("generator_operate", o->x, o->y, 64);
                    register_hook(&self->creating->onremove, &hook_generator_destroy, "a", self);
		}
	    }
	}
    }
}



METHOD(void, generator, draw) {
    if (self->creating) {
	if (self->create_timer <= 8)
	    CALLMETH(self->creating, draw);
	addsprite(self->x, self->y, 0, generator, 12-self->create_timer, DIR_FACE, 0);
    }
}

/*
 * SECRETWALL
 */


ACTOR_ICLASSDEF(secretwall, int nx, px, ny, py; int dir, activated);

SERIALIZER(secretwall) {
    S_INT(nx);
    S_INT(px);
    S_INT(ny);
    S_INT(py);
    S_INT(dir);
    S_INT(activated);
}
METHOD(void, secretwall, init) {
    actor_init(self);

    self->activated = 0;
    self->dir = 0;
    self->w = self->h = 32768;
}

METHOD(void, secretwall, load, buffer* b) {
    actor_load(self, b);
    self->dir = readu8(b);
}

METHOD(void, secretwall, move) {
    if (self->activated == 1) {
        int res = actor_move(self);
        if (res) {
            int bx = self->x>>16;
            int by = self->y>>16;
            self->activated = 2;
            BCELL(bx, by) |= 0xc000;
            BPX(bx - 1, by) = self->nx;
            BNX(bx + 1, by) = self->px;
            BPY(bx, by - 1) = self->ny;
            BNY(bx, by + 1) = self->py;
        }
    }
}

METHOD(int, secretwall, damage, tagset* typ, int amt) {
    if (TAGSET_IN(typ, damages_misc)) {
        if (!self->activated) {
            int bx = self->x>>16;
            int by = self->y>>16;
            self->activated = 1;
            playsndxy("secret_activate",self->x,self->y, 64);
            switch (self->dir) {
            case 0:
                self->vx = 4000;
                break;
            case 1:
                self->vx = -4000;
                break;
            case 2:
                self->vy = 4000;
                break;
            case 3:
                self->vy = -4000;
                break;
            }
            BCELL(bx, by) &= ~0xc000;
            self->nx = BPX(bx - 1, by);
            self->px = BNX(bx + 1, by);
            self->ny = BPY(bx, by - 1);
            self->py = BNY(bx, by + 1);
            BPX(bx - 1, by) = 0;
            BNX(bx + 1, by) = 0;
            BPY(bx, by - 1) = 0;
            BNY(bx, by + 1) = 0;
            return DMG_ABSORB;
        }
    }

    return DMG_NO;
}

METHOD(void, secretwall, draw) {
    if (self->activated == 1) {
        if (self->nx)
            addsprite(self->x-32768, self->y, 0, board, self->nx, DIR_WEST, 0);
        if (self->px)
            addsprite(self->x+32768, self->y, 0, board, self->px, DIR_EAST, 0);
        if (self->ny)
            addsprite(self->x, self->y-32768, 0, board, self->ny, DIR_NORTH, 0);
        if (self->py)
            addsprite(self->x, self->y+32768, 0, board, self->py, DIR_SOUTH, 0);
    }
}

METHOD(int, secretwall, clips) {
    if (self->activated == 1) return 1;
    return 0;
}

METHOD(int, secretwall, pushes) {
    if (self->activated == 1) return 1;
    return 0;
}

/*
 *  KEY
 */

TEXGROUP(key);
ACTOR_ICLASSDEF(key);

METHOD(void, key, draw) {
    addsprite(self->x, self->y, 0, key, 0, DIR_FACE, 0);
}

METHOD(void, key, load, buffer * b) {
    actor_load(self, b);
    self->w = self->h = 16384;
}
METHOD(void, key, touch, actor * o) {
    if (HASTAG(o, is_player)) {
	ADDTAG(o, has_key);
	playsndxy("pickup",self->x,self->y, 64);
	actor_remove(self);
    }
}

/*
 *  MEDKIT
 */

TEXGROUP(medkit);
ACTOR_ICLASSDEF(medkit, int amt);

SERIALIZER(medkit) {
    S_INT(amt);
}

METHOD(void, medkit, draw) {
    if (self->amt)
        addsprite(self->x, self->y, 0, medkit, 0, DIR_FACE, 0);
    else
        addsprite(self->x, self->y, 0, medkit, 1, DIR_FACE, 0);
}

METHOD(void, medkit, load, buffer * b) {
    actor_load(self, b);
    self->w = self->h = 16384;
    self->amt = readu8(b);
}

METHOD(void, medkit, touch, actor* o) {
    if (!self->amt)
        return;
    if (HASTAG(o, is_player)) {
        if (player_add_health(o, self->amt)) {
            playsndxy("pickup",self->x,self->y, 64);
            self->amt = 0;
            /* is something waiting for us to get picked up? */
            if (self->onremove)
                actor_remove(self);
        }
    }
}

/*
 * MAP
 */

static texture maptex;
static int mapvalid;
static int mapflash;
ACTOR_ICLASSDEF(map);

METHOD(void, map, load, buffer* b) {
    actor_load(self, b);

    BCELL(self->x>>16, self->y>>16) = 0xc000;
}
METHOD(void, map, draw) {
    int x, y;
    if (!pointseen(self->x, self->y))
        return;
    if (!maptex.data) {
        maptex.data = malloc(picsize(6));
    }
    if (!mapvalid) {
        mapvalid=1;
        int cx = cplayer->x>>16;
        int cy = cplayer->y>>16;
        mapflash++;
        if (mapflash > 12) mapflash = 0;
        int mf = mapflash;
        if (mf > 6) mf = 13 - mf;
        maptex.shift = 6;
        maptex.sx = 0;
        maptex.ex = 65536;
        for (y = 0; y < 64; y++) {
            for (x = 0; x < 64; x++) {
                int v = 0;
                if (x > 0) v += BPX(x - 1, y);
                if (x < 63) v += BNX(x + 1, y);
                if (y > 0) v += BPY(x, y - 1);
                if (y < 63) v += BNY(x, y + 1);
                if (v) {
                    putpixel(maptex.data, 6, x, y, ((v>>2)&3) + 3);
                }
                else
                    putpixel(maptex.data, 6, x, y, 1);
            }
        }
        putpixel(maptex.data, 6, cx, cy, mf+1);
    }
    addsprite_tex(self->x-32768, self->y, 0, &maptex, DIR_WEST, 0);
    addsprite_tex(self->x+32768, self->y, 0, &maptex, DIR_EAST, 0);
    addsprite_tex(self->x, self->y-32768, 0, &maptex, DIR_NORTH, 0);
    addsprite_tex(self->x, self->y+32768, 0, &maptex, DIR_SOUTH, 0);
    
}

METHOD(void, map, move) {
    mapvalid = 0;
}
