/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#ifndef __MIXER_H_
#define __MIXER_H_

#include "common.h"
#include "class.h"

enum {
    CHANNEL_STARTED = 0,
    CHANNEL_GROUP_FULL = -1,
    CHANNEL_SYS_FULL = -2,
};

typedef struct {
    int freq, channels, bitspersample;
    bool sign, big;
    uint32 internal_format_desc;
} sound_format;

typedef struct {
    sound_format* fmt;
    uint8* buf;
    int nsamps;
} sound_buffer;

typedef struct channel_group {
    int size;
    int num_active;
    struct channel* channels[0];
} channel_group;

typedef struct channelclass {
    CLASS_HEAD;
    void (*halted)(void* self);
    void (*buffer_end)(void* self);
    void (*gametick)(void* self);
    void (*premix)(void* self);
} channelclass;

#define CHANNEL_HEAD                            \
    struct mix_channel* mixchan;                \
    int stamp;                                  \
    channel_group* group;                       \
    sound_format* fmt;

typedef struct channel {
    OBJECT_HEAD(channelclass);
    CHANNEL_HEAD;
} channel;

typedef struct mix_channel {
    /* used by asm - do not rearrange! */
    uint8 *pos, *end;
    uint32 internal_format_desc;
    uint32 lrvol;
    channel* chan;
} mix_channel;

#define CHANNEL_CLASSDEF(nm, ...) _CLASSDEF(channelclass, nm, CHANNEL_HEAD; __VA_ARGS__ );
#define CHANNEL_INITCLASS(nm)                                           \
_INITCLASS(channelclass, nm, channel_init, channel_fini,                \
           (void (*)(void*))_default_method,                            \
           (void (*)(void*))_default_method,                            \
           (void (*)(void*))_default_method,                            \
           (void (*)(void*))_default_method);

#define CHANNEL_ICLASSDEF(nm, ...)          \
CHANNEL_CLASSDEF(nm, ## __VA_ARGS__);       \
CHANNEL_INITCLASS(nm)


void channel_init(void* self);
void channel_fini(void* self);


void mix_init(void);
void mix_fini(void);
void mix_gametick(void);

sound_format* mix_find_format(int freq, int channels, bool big, bool sign, int bitspersample);
int load_wav(const char* fname, sound_buffer* sb);

channel_group* channel_group_new(int size);
void channel_group_free(channel_group*);
void channel_group_halt(channel_group*);
channel* channel_group_oldest(channel_group*);
void channel_set_buffer(channel* self, sound_buffer* buf);
uint8* channel_get_current_pos(channel* self);

int channel_start(channel* chan, channel_group* g, sound_buffer* buf);
void channel_halt(channel* chan);

static inline void channel_set_volume(channel* self, int lvol, int rvol) {

    if (lvol < 0) lvol = 0;
    if (rvol < 0) rvol = 0;

    if (lvol > 32767) lvol = 32767;
    if (rvol > 32767) rvol = 32767;

    self->mixchan->lrvol = ((uint32)lvol)<<16 | ((uint32)rvol);
}

static inline int channel_start_force(channel* chan, channel_group* g, sound_buffer* buf) {
    int ret = channel_start(chan, g, buf);
    if (ret == CHANNEL_GROUP_FULL) {
        channel_halt(channel_group_oldest(g));
        ret = channel_start(chan, g, buf);
    }
    return ret;
}


#endif
