/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#ifndef _UTIL_H_
#define _UTIL_H_

#include "common.h"
#include "datafile.h"

typedef struct {
    char* name;
} namedobject;

typedef struct {
    namedobject** objects;
    int size;
} objectgroup;

typedef struct {
    char* name;
    void (*serial) (buffer* buf, int phase);
} option_handler;

typedef struct {
    uint32 ticks;
    uint16 tickfrac;
} timeprec_t;

#define REGISTER_NAMEDOBJECT(grp,obj)                   \
static void CTOR __register_##grp##_##obj(void) {       \
    _register_namedobject(&grp,&obj);                   \
}

#define _OPTION(name,str)                                       \
static void __option_handler_##name(buffer* buf, int phase);    \
static option_handler option_handler_##name = {                 \
    str, __option_handler_##name                                \
};                                                              \
REGISTER_NAMEDOBJECT(option_handlers, option_handler_##name);   \
static void __option_handler_##name(buffer* buf, int phase)

#define OPTION(name) _OPTION(name, #name)
#define OPTION_INT(name) _OPTION(name, #name) { serial_u32(&name, buf, phase); }

extern objectgroup option_handlers;

#ifdef SDLTEST
extern char lcd_framebuffer[5120];
#endif


#ifdef SDL_TEST
#define get_time_div(tt) 		        \
({			                        \
    (tt).ticks = SDL_GetTicks();                \
})


#elif defined(CPU_COLDFIRE)

#define get_time_div(tt) 			\
({						\
    uint16 cn0 = TCN0;				\
    uint16 tk0 = *RB current_tick;		\
    uint16 cn1 = TCN0;				\
    uint16 tk1 = *RB current_tick;		\
 						\
						\
    if (tk0 == tk1) {				\
	(tt).tickfrac = cn1;			\
	(tt).ticks = tk1;			\
    } else {					\
	(tt).tickfrac = cn0;			\
	(tt).ticks = tk0;			\
    }						\
})

#else

#define get_time_div(tt)                        \
({                                              \
    (tt).ticks = *RB current_tick;              \
    (tt).tickfrac = 0;                          \
})

#endif

#define SCREEN_SIZE sizeof(lcd_framebuffer)

int load_settings(const char* path);
int save_settings(const char* path);

void _register_namedobject(objectgroup* grp, void* obj);
namedobject* find_namedobject(objectgroup* grp, const char* name);
namedobject* read_namedobject(objectgroup* grp, buffer* b);

uint32 get_interval_usec(timeprec_t* start, timeprec_t* end);
void* save_screen();
void restore_screen(void* buffer);

#endif
